;;; Autor: Jose Ignacio Aguillo Mendi
;;; E-mail: pantic@infonegocio.com
;;; E-mail2: 452446@cepsz.unizar.es
;;; Fichero: ierl21.lsp
;;; Contenido: Implementacin de las funcionalidades agnadidas a la version IERL 2.0
;;; Fecha de ultima modificacion: 20-01-2002
;;; Comentarios: El presente fichero se plantea como una especie de parche que define nuevas funciones
;;; y redefine otras que ya existian en la version 2.0. Por tanto, es necesario cargarlo tras haber cargado
;;; los ficheros: ierl Forward Chaining.lsp, ierl Pattern Matching.lsp, ierl Rule Definition.lsp y ierl Streams.lsp.
;;; 

;;; CAMBIOS Y AADIDOS EN LA CORRESPONDENCIA DE PATRONES
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; "resetea" el sistema experto
(defun clear ()
  (setf *agenda* nil)
  (setf *reglas* 'flujo-vacio)
  (setf *afirmaciones* 'flujo-vacio)
  (setf *frames* 'flujo-vacio))

;;; permite ver los hechos de memoria de trabajo en un momento determinado
(defun ver-afirmaciones ()
  (ver-afirmaciones-aux *afirmaciones*))

(defun ver-afirmaciones-aux (flujo-afirmaciones)
  (cond
   ((final-del-flujo-p flujo-afirmaciones) 'NO-MAS)
   (t 
    (format t "~%~a" (principio-del-flujo flujo-afirmaciones))
    (ver-afirmaciones-aux (resto-del-flujo flujo-afirmaciones)))))

;;; genera las posibles combinaciones de valores de una lista para una variable de tipo lista
(defun genera-posibilidades (lista)
  (genera-posibilidades-aux nil lista))

(defun genera-posibilidades-aux (primero resto)
  (cond
   ((endp resto) (list (list primero)))
   (t (cons (cons primero resto) (genera-posibilidades-aux (append primero (list (first resto))) (rest resto))))))

;;; identifica la aparicin en el patrn de una variable de tipo lista
(defun variable-$-p (patron)
  (and (listp patron)
       (listp (first patron))
       (eq (first (first patron)) '$)))

;;; Recordar: ($ x) se sustituye por los elementos de la lista y (? x) por la lista
(defun FC-particulariza-variables (patron lista-a)
  (cond   ((atom patron) patron)
          ((eq '? (first patron))
           (extrae-valor (encuentra-ligadura patron lista-a)))
          ((eq 'objeto (first patron))
           (extrae-valor (encuentra-ligadura patron lista-a)))
          ((eq '$ (first patron))
           (extrae-valor (encuentra-ligadura patron lista-a)))
          ((variable-$-p patron)
           (append (FC-particulariza-variables (first patron) lista-a)
                   (FC-particulariza-variables (rest patron) lista-a)))
          (t (cons (FC-particulariza-variables (first patron) lista-a)
                   (FC-particulariza-variables (rest patron) lista-a)))))

;;; identifica una lista formada por listas de ligaduras y hechos
(defun es-lista-de-listas-lig-y-hechos? (lista)
  (and (listp (car lista))
       (listp (caar lista))
       (listp (caaar lista))
       (not (eq nil (caaar lista)))))

;;; identifica una lista de listas de ligaduras
(defun es-lista-de-listas-lig? (lista)
  (and (listp (car lista))
       (listp (caar lista))
       (not (eq nil (caar lista)))))

;;;ligaduras-y-hechos=(lista de ligaduras lista de hechos)
(defun corresponde-todos-hechos (patron dato &optional ligaduras-y-hechos)
   (cond
    ((eq '<- (second patron)) ;;; es una asignacion de hecho a una variable
      (let ((res (corresponde (third patron) dato (first ligaduras-y-hechos))))
         (if (eq res 'FALLA)
            'FALLA
            (if (es-lista-de-listas-lig? res)
               (do* ((resultado res (rest resultado))
                     (ligs (first resultado) (first resultado))
                     (devuelve nil))
                    ((endp resultado) devuelve)
                  (setf devuelve (append devuelve (list (list (agrega-ligadura (first patron) dato ligs)
                                                              (append (second ligaduras-y-hechos) 
                                                                (if (member dato (second ligaduras-y-hechos) :test #'equal)
                                                                   nil
                                                                   (list dato))))))))
               (list (agrega-ligadura (first patron) dato res)
                     (append (second ligaduras-y-hechos) 
                       (if (member dato (second ligaduras-y-hechos) :test #'equal)
                          nil
                          (list dato))))))))
    (t   ;;; es un antecedente normal
      (let ((res (corresponde patron dato (first ligaduras-y-hechos))))
         (if (eq res 'FALLA)
            'FALLA
            (if (es-lista-de-listas-lig? res)
               (do* ((resultado res (rest resultado))
                     (ligs (first resultado) (first resultado))
                     (devuelve nil))
                    ((endp resultado) devuelve)
                  (setf devuelve (append devuelve (list (list ligs (append (second ligaduras-y-hechos) 
                                                                     (if (member dato (second ligaduras-y-hechos) :test #'equal)
                                                                        nil
                                                                        (list dato))))))))
               (list res (append (second ligaduras-y-hechos) 
                           (if (member dato (second ligaduras-y-hechos) :test #'equal)
                              nil
                              (list dato))))))))))

;;;Se asume que nil no puede aparecer en los hechos ej: (a b nil c d)
(defun corresponde (patron dato &optional ligaduras)
  (cond   ((and (eq patron nil) (eq dato nil)) ligaduras)
          ((elementos-p patron dato)
           (corresponde-atomos patron dato ligaduras))
          ((variable-p patron)
           (corresponde-variable patron dato ligaduras))
          ((variable-$-p patron)
           (corresponde-con-$ patron dato ligaduras))
          ((recursivos-p patron dato)
           (corresponde-partes patron dato ligaduras))
          (t
           'FALLA)))

(defun corresponde-con-$ (patron dato ligaduras)
    (let ((ligadura (encuentra-ligadura (first patron) ligaduras)))
      (if ligadura
          (corresponde (append (extrae-valor ligadura) (rest patron)) dato ligaduras) ;;;sustituir por el valor de la ligadura
        (let ((posibilidades (genera-posibilidades dato)))              
          (do* ((posibilidades posibilidades (rest posibilidades))
                (resultado nil)
                (bandera-exito nil))
               ((endp posibilidades) (if (eq bandera-exito nil)
                                         'FALLA   ; falla con todas las posibilidades
                                       resultado))
            (if (predicados-satisfechos-p (extrae-predicados (first patron)) (caar posibilidades))
                (let ((res (corresponde (rest patron) (rest (first posibilidades)) (agrega-ligadura (first patron) (caar posibilidades) ligaduras))))
                  (if (not (eq res 'FALLA))
                      (progn
                        (setf bandera-exito t)
                      (if (es-lista-de-listas-lig? res)
                          (setf resultado (append
                                       resultado
                                           res))
                        (setf resultado (append resultado (list res))))))
                  )
              ))))))

(defun corresponde-partes (patron dato ligaduras)
    (let ((resultado (corresponde (car patron) (car dato) ligaduras)))
        (if (eq 'FALLA resultado)
            'FALLA
                (if (es-lista-de-listas-lig? resultado) ;;;es una lista de de listas de ligaduras 
                    (do* ((resultado resultado (rest resultado))
                          (lig (first resultado) (first resultado))
                          (res nil)
                          (bandera-exito nil))
                         ((endp resultado) (if (eq bandera-exito nil)
                                               'FALLA  ;falla con todas las posibilidades
                                             res))
                      (let ((result (corresponde (cdr patron) (cdr dato) lig)))
                        (if (not (eq result 'FALLA))
                            (progn
                              (setf bandera-exito t)
                              (if (es-lista-de-listas-lig? result)
                                  (setf res (append res result))
                                (setf res (append res (list result))))))))
                  (corresponde (cdr patron) (cdr dato) resultado)))
            ))
                      
;Se asume que la correspondencia entre una variable y nil falla siempre                                            
(defun corresponde-variable (patron dato ligaduras)
  (if (eq dato nil)
      'FALLA
   (if (predicados-satisfechos-p (extrae-predicados patron) dato)
       (let ((ligadura (encuentra-ligadura patron ligaduras)))
           (if ligadura
               (corresponde (extrae-valor ligadura) dato ligaduras)
                   (agrega-ligadura patron dato ligaduras)))
     'FALLA)))


(defun corresponde-atomos (patron dato ligaduras)
  (if (eq patron dato)
          ligaduras
    'FALLA))

(defun intenta-afirmacion (patron afirmacion ligaduras-y-hechos)
    (let ((resultado (corresponde-todos-hechos patron afirmacion ligaduras-y-hechos)))
        (if (eq 'FALLA resultado)
            'FLUJO-VACIO
          (if (es-lista-de-listas-lig-y-hechos? resultado)
              (construir-un-flujo resultado)
            (construye-flujo resultado 'FLUJO-VACIO)))))

(defun construir-un-flujo (lista-de-listas)
  (cond
   ((endp lista-de-listas) 'flujo-vacio)
   (t (let ((primero (first lista-de-listas)))
        (construye-flujo primero (construir-un-flujo (rest lista-de-listas)))))))

;;; FUNCIONES PARA LA GESTIN DE LA AGENDA
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; variable global en la que se implementa la agenda
(defvar *agenda* nil)

;;; introduce una regla en la agenda
;;; una regla en la agenda se representa como una lista con su nombre, los hechos que
;;; la sensibilizan, las ligaduras que dan lugar a esos hechos y un valor que indica si
;;; la regla ha sido disparada o no
(defun agrega-regla-a-agenda (nombre-regla hechos ligaduras disparada?)
  (setf *agenda* (append (list (list nombre-regla hechos ligaduras disparada?)) *agenda*)))

;;; primera regla de la agenda
(defun prim-reg-ag (agenda)
  (first agenda))

;;; resto de reglas de la agenda
(defun resto-reg-ag (agenda)
  (rest agenda))

(defun nombre-regla-ag (regla)
  (first regla))

(defun hechos-regla-ag (regla)
  (second regla))

(defun ligaduras-regla-ag (regla)
  (third regla))

(defun disparada-regla-ag? (regla)
  (fourth regla))


;;; encuentra en la agenda el disparo (hechos) si existe
(defun encuentra-disparo (regla disparo)
  (encuentra-disparo-aux regla disparo *agenda*))

(defun encuentra-disparo-aux (regla disparo agenda)
  (cond
   ((endp agenda) nil)
   ((and (eq (nombre-de-la-regla regla) (nombre-regla-ag (first agenda)))
         (equal disparo (hechos-regla-ag (first agenda))))
    (first agenda))
   (t (encuentra-disparo-aux regla disparo (rest agenda)))))

;;; intenta meter los posibles disparos de la regla en la agenda
(defun intenta-regla (regla)
  (let ((flujo-ligaduras
         (fc-aplica-filtros (antecedentes-de-la-regla regla)
                         (construye-flujo nil 'flujo-vacio))))
    (do ((flujo-ligaduras flujo-ligaduras (resto-del-flujo flujo-ligaduras)))
        ((final-del-flujo-p flujo-ligaduras) t)
      (if (not (encuentra-disparo regla (second (principio-del-flujo flujo-ligaduras))))
          (agrega-regla-a-agenda (nombre-de-la-regla regla) ; el nombre
                                 (second (principio-del-flujo flujo-ligaduras)) ;los hechos
                                 (first (principio-del-flujo flujo-ligaduras)) ;las ligaduras
                                 nil) ; no esta disparada
        ))))

;;; intenta-regla con todas las reglas
(defun prueba-reglas ()
    (do ((flujo-de-reglas *reglas* (resto-del-flujo flujo-de-reglas)))
        ((final-del-flujo-p flujo-de-reglas))
      (intenta-regla (principio-del-flujo flujo-de-reglas))))
  
;;; Se asume que las reglas que no han sido disparadas estan al principio de la agenda y las disparadas al final 
(defun agenda-vacia? ()
  (disparada-regla-ag? (prim-reg-ag *agenda*)))

;;; obtiene la regla del flujo de reglas a partir de su nombre obtenido en la agenda
(defun obten-regla-del-flujo (nombre-regla)
  (do ((flujo-de-reglas *reglas* (resto-del-flujo flujo-de-reglas))
       (regla nil))
      ((or (final-del-flujo-p flujo-de-reglas) (not (endp regla))) regla)
    (if (eq nombre-regla (nombre-de-la-regla (principio-del-flujo flujo-de-reglas)))
        (setf regla (principio-del-flujo flujo-de-reglas)))))
    
;;; dispara la primera regla de la agenda
(defun dispara-primera-regla ()
  (let* ((regla (obten-regla-del-flujo (nombre-regla-ag (prim-reg-ag *agenda*))))
         (primera-regla (prim-reg-ag *agenda*))
         (resultado  (FC-particulariza-variables
                     (consecuente-de-la-regla regla)
                     (ligaduras-regla-ag primera-regla))))
    (if *ver-reglas* (format t "~&Regla ~a disparada con los hechos: ~a." (nombre-de-la-regla regla)(hechos-regla-ag primera-regla)))
    (eval resultado)
    (setf *agenda* (append (resto-reg-ag *agenda*)
                           (list (list (nombre-regla-ag primera-regla) (hechos-regla-ag primera-regla)
                                       (ligaduras-regla-ag primera-regla) t))))
     ))
    
;;; controla que la regla no haya sido disparada ya (refraccion)
(defun funcion-de-ordenacion (regla1 regla2)
  (and (not (disparada-regla-ag? regla1))
       (estrategia-definida regla1 regla2)))

;;; implementacin de la estrategia de control MEA
(defun estrategia-definida (regla1 regla2)
   (if (> (recencia (first (hechos-regla-ag regla1))) (recencia (first (hechos-regla-ag regla2))))
      t
      (if (= (recencia (first (hechos-regla-ag regla1))) (recencia (first (hechos-regla-ag regla2))))
       (> (length (hechos-regla-ag regla1)) (length (hechos-regla-ag regla2)))
         nil)))

;;; ordena las reglas en la agenda
(defun ordena-reglas ()
    (setf *agenda* (sort *agenda* #'funcion-de-ordenacion)))

;;;Funcion de encadenamiento progresivo
(defun encadenamiento-progresivo ()
   (prueba-reglas)
   (do ()
       ((agenda-vacia?) 'HECHO)
           (ordena-reglas)
           (dispara-primera-regla)
           (prueba-reglas)))

;;; devuelve la agenda borrando los disparos que contengan el hecho
(defun quita-disparos-de-agenda (hecho agenda)
  (cond
   ((endp agenda) nil)
   ((member hecho (hechos-regla-ag (prim-reg-ag agenda)) :test #'equal)
    (quita-disparos-de-agenda hecho (resto-reg-ag agenda)))
   (t
    (cons (prim-reg-ag agenda) (quita-disparos-de-agenda hecho (resto-reg-ag agenda))))))

;;; variables globales que controlan el comportamiento de watch y unwatch
(defvar *ver-afirmaciones* nil)
(defvar *ver-reglas* nil)

;;; permite ver la insercion/retirada de hechos y/o disparos de reglas
(defun watch (&rest opciones)
   (dolist (opcion opciones t)
      (cond
       ((eq opcion 'afirmaciones) (setf *ver-afirmaciones* t))
       ((eq opcion 'reglas) (setf *ver-reglas* t)))))

;;; desactiva el comportamiento definido por watch
(defun unwatch (&rest opciones)
   (dolist (opcion opciones t)
      (cond
       ((eq opcion 'afirmaciones) (setf *ver-afirmaciones* nil))
       ((eq opcion 'reglas) (setf *ver-reglas* nil)))))

;;; calcula la recencia de un hecho, ser mayor contra ms se acerque al final del flujo de afirmaciones
(defun recencia (hecho)
   (do ((suma 0 (+ 1 suma))
        (flujo-hechos *afirmaciones* (resto-del-flujo flujo-hechos)))
       ((equal hecho (principio-del-flujo flujo-hechos)) (+ 1 suma))))

;;; permite ver el estado de la agenda
(defun ver-agenda ()
   (eval '*agenda*))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; FUNCIONES PARA EL MANEJO DE HECHOS EN MEMORIA DE TRABAJO
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Funcin assert!, introduce un hecho en memoria de trabajo
(defun assert! (hecho)
  (when (recuerda-afirmacion hecho)
    (if *ver-afirmaciones* (format t "~&=> ~a" hecho))))

;;; Funcion retract!, retira un hecho de memoria de trabajo
(defun retract! (hecho)
  (if (miembro-de-flujo hecho *afirmaciones*)
      (progn 
        (if *ver-afirmaciones* (format t "~&<= ~a" hecho))
        (borra-de-flujo hecho *afirmaciones*)
        (setf *agenda* (quita-disparos-de-agenda hecho *agenda*)))
    nil
  ))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;FUNCIONES AADIDAS PARA LA GESTIN DE FLUJOS
(defmacro borra-de-flujo (objeto flujo)
  `(setf ,flujo
    (borra-de-flujo-aux ,objeto ,flujo)))

(defun borra-de-flujo-aux (objeto flujo)
  (cond
   ((final-del-flujo-p flujo) 'flujo-vacio)
   ((equal objeto (principio-del-flujo flujo)) (resto-del-flujo flujo))
   (t (construye-flujo (principio-del-flujo flujo) (borra-de-flujo-aux objeto (resto-del-flujo flujo))))))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
